# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from importlib.metadata import version as get_version

project = "WuttaWeb"
copyright = "2024, Lance Edgar"
author = "Lance Edgar"
release = get_version("WuttaWeb")

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "sphinxcontrib.programoutput",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

intersphinx_mapping = {
    "colander": ("https://docs.pylonsproject.org/projects/colander/en/latest/", None),
    "deform": ("https://docs.pylonsproject.org/projects/deform/en/latest/", None),
    "fanstatic": ("https://www.fanstatic.org/en/latest/", None),
    "pyramid": ("https://docs.pylonsproject.org/projects/pyramid/en/latest/", None),
    "python": ("https://docs.python.org/3/", None),
    "rattail-manual": ("https://docs.wuttaproject.org/rattail-manual/", None),
    "sqlalchemy": ("http://docs.sqlalchemy.org/en/latest/", None),
    "webhelpers2": ("https://webhelpers2.readthedocs.io/en/latest/", None),
    "wuttjamaican": ("https://docs.wuttaproject.org/wuttjamaican/", None),
    "wutta-continuum": ("https://docs.wuttaproject.org/wutta-continuum/", None),
}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for people
"""

import sqlalchemy as sa

from wuttjamaican.db.model import Person
from wuttaweb.views import MasterView
from wuttaweb.util import make_users_grid


class PersonView(MasterView):  # pylint: disable=abstract-method
    """
    Master view for people.

    Default route prefix is ``people``.

    Notable URLs provided by this class:

    * ``/people/``
    * ``/people/new``
    * ``/people/XXX``
    * ``/people/XXX/edit``
    * ``/people/XXX/delete``
    """

    model_class = Person
    model_title_plural = "People"
    route_prefix = "people"
    sort_defaults = "full_name"
    has_autocomplete = True

    grid_columns = [
        "full_name",
        "first_name",
        "middle_name",
        "last_name",
    ]

    filter_defaults = {
        "full_name": {"active": True},
    }

    form_fields = [
        "full_name",
        "first_name",
        "middle_name",
        "last_name",
        "users",
    ]

    def configure_grid(self, grid):  # pylint: disable=empty-docstring
        """ """
        g = grid
        super().configure_grid(g)

        # full_name
        g.set_link("full_name")

        # first_name
        g.set_link("first_name")

        # last_name
        g.set_link("last_name")

    def configure_form(self, form):  # pylint: disable=empty-docstring
        """ """
        f = form
        super().configure_form(f)
        person = f.model_instance

        # full_name
        if self.creating or self.editing:
            f.remove("full_name")

        # users
        if self.viewing:
            f.set_grid("users", self.make_users_grid(person))

    def make_users_grid(self, person):
        """
        Make and return the grid for the Users field.

        This grid is shown for the Users field when viewing a Person.

        :returns: Fully configured :class:`~wuttaweb.grids.base.Grid`
           instance.
        """
        return make_users_grid(
            self.request,
            route_prefix=self.get_route_prefix(),
            data=person.users,
            columns=[
                "username",
                "active",
            ],
        )

    def objectify(self, form):  # pylint: disable=empty-docstring
        """ """
        person = super().objectify(form)

        # full_name
        person.full_name = self.app.make_full_name(person.first_name, person.last_name)

        return person

    def autocomplete_query(self, term):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        session = self.Session()
        query = session.query(model.Person)
        criteria = [model.Person.full_name.ilike(f"%{word}%") for word in term.split()]
        query = query.filter(sa.and_(*criteria)).order_by(model.Person.full_name)
        return query

    def view_profile(self, session=None):  # pylint: disable=empty-docstring
        """ """
        person = self.get_instance(session=session)
        context = {
            "person": person,
            "instance": person,
        }
        return self.render_to_response("view_profile", context)

    def make_user(self):  # pylint: disable=empty-docstring
        """ """
        self.request.session.flash("TODO: this feature is not yet supported", "error")
        return self.redirect(self.request.get_referrer())

    @classmethod
    def defaults(cls, config):  # pylint: disable=empty-docstring
        """ """

        # nb. Person may come from custom model
        wutta_config = config.registry.settings["wutta_config"]
        app = wutta_config.get_app()
        cls.model_class = app.model.Person

        cls._defaults(config)
        cls._people_defaults(config)

    @classmethod
    def _people_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()
        instance_url_prefix = cls.get_instance_url_prefix()
        permission_prefix = cls.get_permission_prefix()

        # view profile
        config.add_route(
            f"{route_prefix}.view_profile",
            f"{instance_url_prefix}/profile",
            request_method="GET",
        )
        config.add_view(
            cls,
            attr="view_profile",
            route_name=f"{route_prefix}.view_profile",
            permission=f"{permission_prefix}.view_profile",
        )

        # make user for person
        config.add_route(
            f"{route_prefix}.make_user",
            f"{url_prefix}/make-user",
            request_method="POST",
        )
        config.add_view(
            cls,
            attr="make_user",
            route_name=f"{route_prefix}.make_user",
            permission="users.create",
        )


def defaults(config, **kwargs):  # pylint: disable=missing-function-docstring
    base = globals()

    PersonView = kwargs.get(  # pylint: disable=invalid-name,redefined-outer-name
        "PersonView", base["PersonView"]
    )
    PersonView.defaults(config)


def includeme(config):  # pylint: disable=missing-function-docstring
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaWeb - test utilities
"""

import sys
from unittest.mock import MagicMock

import fanstatic
import pytest
from pyramid import testing

from wuttjamaican.testing import DataTestCase
from wuttjamaican.db.model.base import metadata

from wuttaweb import subscribers
from wuttaweb.conf import WuttaWebConfigExtension


class WebTestCase(DataTestCase):
    """
    Base class for test suites requiring a full (typical) web app.
    """

    def setUp(self):  # pylint: disable=empty-docstring
        """ """
        self.setup_web()

    def setup_web(self):
        """
        Perform setup for the testing web app.
        """
        self.setup_db()
        self.request = self.make_request()
        self.pyramid_config = testing.setUp(
            request=self.request,
            settings={
                "wutta_config": self.config,
                "mako.directories": ["wuttaweb:templates"],
                "pyramid_deform.template_search_path": "wuttaweb:templates/deform",
            },
        )

        # init web
        self.pyramid_config.include("pyramid_deform")
        self.pyramid_config.include("pyramid_mako")
        self.pyramid_config.add_directive(
            "add_wutta_permission_group", "wuttaweb.auth.add_permission_group"
        )
        self.pyramid_config.add_directive(
            "add_wutta_permission", "wuttaweb.auth.add_permission"
        )
        self.pyramid_config.add_subscriber(
            "wuttaweb.subscribers.before_render", "pyramid.events.BeforeRender"
        )
        self.pyramid_config.include("wuttaweb.static")

        # nb. mock out fanstatic env..good enough for now to avoid errors..
        needed = fanstatic.init_needed()
        self.request.environ[fanstatic.NEEDED] = needed

        # setup new request w/ anonymous user
        event = MagicMock(request=self.request)
        subscribers.new_request(event)

        def user_getter(request, **kwargs):  # pylint: disable=unused-argument
            pass

        subscribers.new_request_set_user(
            event, db_session=self.session, user_getter=user_getter
        )

    def tearDown(self):
        self.teardown_web()

    def teardown_web(self):
        """
        Perform teardown for the testing web app.
        """
        testing.tearDown()
        self.teardown_db()

    def make_request(self):
        """
        Make and return a new dummy request object.
        """
        return testing.DummyRequest(client_addr="127.0.0.1")


@pytest.mark.versioned
class VersionWebTestCase(WebTestCase):
    """
    Base class for test suites requiring a full (typical) web app,
    with Continuum versioning support.
    """

    def setUp(self):
        self.setup_versioning()

    def setup_versioning(self):
        """
        Perform setup for the testing web app.
        """
        self.setup_web()

    def tearDown(self):
        self.teardown_versioning()

    def teardown_versioning(self):
        """
        Perform teardown for the testing web app.
        """
        import sqlalchemy_continuum as continuum  # pylint: disable=import-outside-toplevel

        continuum.remove_versioning()
        continuum.versioning_manager.transaction_cls = continuum.TransactionFactory()
        self.teardown_web()

    def make_config(self, **kwargs):
        """
        Make and customize the config object.

        We override this to explicitly enable the versioning feature.
        """
        from wutta_continuum.conf import (  # pylint: disable=import-outside-toplevel
            WuttaContinuumConfigExtension,
        )

        config = super().make_config(**kwargs)
        config.setdefault("wutta_continuum.enable_versioning", "true")

        # nb. must purge model classes from sys.modules, so they will
        # be reloaded and sqlalchemy-continuum can reconfigure
        if "wuttjamaican.db.model" in sys.modules:
            del sys.modules["wuttjamaican.db.model.batch"]
            del sys.modules["wuttjamaican.db.model.upgrades"]
            del sys.modules["wuttjamaican.db.model.auth"]
            del sys.modules["wuttjamaican.db.model.base"]
            del sys.modules["wuttjamaican.db.model"]

        self.assertNotIn("user_version", metadata.tables)

        ext = WuttaWebConfigExtension()
        ext.configure(config)

        ext = WuttaContinuumConfigExtension()
        ext.startup(config)

        return config



# -*- coding: utf-8; -*-

from pyramid import testing

from wuttaweb.views import progress as mod
from wuttaweb.progress import get_progress_session
from wuttaweb.testing import WebTestCase


class TestProgressView(WebTestCase):

    def test_includeme(self):
        self.pyramid_config.include("wuttaweb.views.progress")

    def test_basic(self):
        self.request.session.id = "mockid"
        self.request.matchdict = {"key": "foo"}

        # first call with no setup, will create the progress session
        # but it should be "empty" - except not really since beaker
        # adds some keys by default
        context = mod.progress(self.request)
        self.assertIsInstance(context, dict)

        # now let's establish a progress session of our own
        progsess = get_progress_session(self.request, "bar")
        progsess["maximum"] = 2
        progsess["value"] = 1
        progsess.save()

        # then call view, check results
        self.request.matchdict = {"key": "bar"}
        context = mod.progress(self.request)
        self.assertEqual(context["maximum"], 2)
        self.assertEqual(context["value"], 1)
        self.assertNotIn("complete", context)

        # now mark it as complete, check results
        progsess["complete"] = True
        progsess["success_msg"] = "yay!"
        progsess.save()
        context = mod.progress(self.request)
        self.assertTrue(context["complete"])
        self.assertEqual(context["success_msg"], "yay!")

        # now do that all again, with error
        progsess = get_progress_session(self.request, "baz")
        progsess["maximum"] = 2
        progsess["value"] = 1
        progsess.save()
        self.request.matchdict = {"key": "baz"}
        context = mod.progress(self.request)
        self.assertEqual(context["maximum"], 2)
        self.assertEqual(context["value"], 1)
        self.assertNotIn("complete", context)
        self.assertNotIn("error", context)
        progsess["error"] = True
        progsess["error_msg"] = "omg!"
        progsess.save()
        context = mod.progress(self.request)
        self.assertTrue(context["error"])
        self.assertEqual(context["error_msg"], "omg!")



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for app settings
"""

import datetime
import json
import os
import sys
import subprocess
from collections import OrderedDict

from wuttjamaican.db.model import Setting
from wuttjamaican.util import get_timezone_by_name
from wuttaweb.views import MasterView
from wuttaweb.util import get_libver, get_liburl


class AppInfoView(MasterView):  # pylint: disable=abstract-method
    """
    Master view for the core app info, to show/edit config etc.

    Default route prefix is ``appinfo``.

    Notable URLs provided by this class:

    * ``/appinfo/``
    * ``/appinfo/configure``

    See also :class:`SettingView`.
    """

    model_name = "AppInfo"
    model_title_plural = "App Info"
    route_prefix = "appinfo"
    filterable = False
    sort_on_backend = False
    sort_defaults = "name"
    paginated = False
    creatable = False
    viewable = False
    editable = False
    deletable = False
    configurable = True

    grid_columns = [
        "name",
        "version",
        "editable_project_location",
    ]

    # TODO: for tailbone backward compat with get_liburl() etc.
    weblib_config_prefix = None

    def get_grid_data(  # pylint: disable=empty-docstring
        self, columns=None, session=None
    ):
        """ """

        # nb. init with empty data, only load it upon user request
        if not self.request.GET.get("partial"):
            return []

        # TODO: pretty sure this is not cross-platform.  probably some
        # sort of pip methods belong on the app handler?  or it should
        # have a pip handler for all that?
        pip = os.path.join(sys.prefix, "bin", "pip")
        output = subprocess.check_output([pip, "list", "--format=json"], text=True)
        data = json.loads(output.strip())

        # must avoid null values for sort to work right
        for pkg in data:
            pkg.setdefault("editable_project_location", "")

        return data

    def configure_grid(self, grid):  # pylint: disable=empty-docstring
        """ """
        g = grid
        super().configure_grid(g)

        g.sort_multiple = False

        # name
        g.set_searchable("name")

        # editable_project_location
        g.set_searchable("editable_project_location")

    def get_weblibs(self):  # pylint: disable=empty-docstring
        """ """
        return OrderedDict(
            [
                ("vue", "(Vue2) Vue"),
                ("vue_resource", "(Vue2) vue-resource"),
                ("buefy", "(Vue2) Buefy"),
                ("buefy.css", "(Vue2) Buefy CSS"),
                ("fontawesome", "(Vue2) FontAwesome"),
                ("bb_vue", "(Vue3) vue"),
                ("bb_oruga", "(Vue3) @oruga-ui/oruga-next"),
                ("bb_oruga_bulma", "(Vue3) @oruga-ui/theme-bulma (JS)"),
                ("bb_oruga_bulma_css", "(Vue3) @oruga-ui/theme-bulma (CSS)"),
                ("bb_fontawesome_svg_core", "(Vue3) @fortawesome/fontawesome-svg-core"),
                ("bb_free_solid_svg_icons", "(Vue3) @fortawesome/free-solid-svg-icons"),
                ("bb_vue_fontawesome", "(Vue3) @fortawesome/vue-fontawesome"),
            ]
        )

    def configure_get_simple_settings(self):  # pylint: disable=empty-docstring
        """ """
        simple_settings = [
            # basics
            {"name": f"{self.config.appname}.app_title"},
            {"name": f"{self.config.appname}.node_type"},
            {"name": f"{self.config.appname}.node_title"},
            {"name": f"{self.config.appname}.production", "type": bool},
            {"name": "wuttaweb.themes.expose_picker", "type": bool},
            {"name": f"{self.config.appname}.timezone.default"},
            {"name": f"{self.config.appname}.web.menus.handler.spec"},
            # nb. this is deprecated; we define so it is auto-deleted
            # when we replace with newer setting
            {"name": f"{self.config.appname}.web.menus.handler_spec"},
            # user/auth
            {"name": "wuttaweb.home_redirect_to_login", "type": bool, "default": False},
            # email
            {
                "name": f"{self.config.appname}.mail.send_emails",
                "type": bool,
                "default": False,
            },
            {"name": f"{self.config.appname}.email.default.sender"},
            {"name": f"{self.config.appname}.email.default.subject"},
            {"name": f"{self.config.appname}.email.default.to"},
            {"name": f"{self.config.appname}.email.feedback.subject"},
            {"name": f"{self.config.appname}.email.feedback.to"},
        ]

        def getval(key):
            return self.config.get(f"wuttaweb.{key}")

        weblibs = self.get_weblibs()
        for key in weblibs:

            simple_settings.append(
                {
                    "name": f"wuttaweb.libver.{key}",
                    "default": getval(f"libver.{key}"),
                }
            )
            simple_settings.append(
                {
                    "name": f"wuttaweb.liburl.{key}",
                    "default": getval(f"liburl.{key}"),
                }
            )

        return simple_settings

    def configure_check_timezone(self):
        """
        AJAX view to validate a user-specified timezone name.

        Route name for this is: ``appinfo.check_timezone``
        """
        tzname = self.request.GET.get("tzname")
        if not tzname:
            return {"invalid": "Must provide 'tzname' parameter."}
        try:
            get_timezone_by_name(tzname)
            return {"invalid": False}
        except Exception as err:  # pylint: disable=broad-exception-caught
            return {"invalid": str(err)}

    def configure_get_context(  # pylint: disable=empty-docstring,arguments-differ
        self, **kwargs
    ):
        """ """
        context = super().configure_get_context(**kwargs)

        # default system timezone
        dt = datetime.datetime.now().astimezone()
        context["default_timezone"] = dt.tzname()

        # add registered menu handlers
        web = self.app.get_web_handler()
        handlers = web.get_menu_handler_specs()
        handlers = [{"spec": spec} for spec in handlers]
        context["menu_handlers"] = handlers

        # add `weblibs` to context, based on config values
        weblibs = self.get_weblibs()
        for key in weblibs:
            title = weblibs[key]
            weblibs[key] = {
                "key": key,
                "title": title,
                # nb. these values are exactly as configured, and are
                # used for editing the settings
                "configured_version": get_libver(
                    self.request,
                    key,
                    prefix=self.weblib_config_prefix,
                    configured_only=True,
                ),
                "configured_url": get_liburl(
                    self.request,
                    key,
                    prefix=self.weblib_config_prefix,
                    configured_only=True,
                ),
                # nb. these are for display only
                "default_version": get_libver(
                    self.request,
                    key,
                    prefix=self.weblib_config_prefix,
                    default_only=True,
                ),
                "live_url": get_liburl(
                    self.request, key, prefix=self.weblib_config_prefix
                ),
            }
        context["weblibs"] = list(weblibs.values())

        return context

    @classmethod
    def defaults(cls, config):  # pylint: disable=empty-docstring
        """ """
        cls._defaults(config)
        cls._appinfo_defaults(config)

    @classmethod
    def _appinfo_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        url_prefix = cls.get_url_prefix()

        # check timezone
        config.add_route(
            f"{route_prefix}.check_timezone",
            f"{url_prefix}/check-timezone",
            request_method="GET",
        )
        config.add_view(
            cls,
            attr="configure_check_timezone",
            route_name=f"{route_prefix}.check_timezone",
            permission=f"{permission_prefix}.configure",
            renderer="json",
        )


class SettingView(MasterView):  # pylint: disable=abstract-method
    """
    Master view for the "raw" settings table.

    Default route prefix is ``settings``.

    Notable URLs provided by this class:

    * ``/settings/``

    See also :class:`AppInfoView`.
    """

    model_class = Setting
    model_title = "Raw Setting"
    deletable_bulk = True
    filter_defaults = {
        "name": {"active": True},
    }
    sort_defaults = "name"

    # TODO: master should handle this (per model key)
    def configure_grid(self, grid):  # pylint: disable=empty-docstring
        """ """
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def configure_form(self, form):  # pylint: disable=empty-docstring
        """ """
        f = form
        super().configure_form(f)

        # name
        f.set_validator("name", self.unique_name)

        # value
        # TODO: master should handle this (per column nullable)
        f.set_required("value", False)

    def unique_name(self, node, value):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        session = self.Session()

        query = session.query(model.Setting).filter(model.Setting.name == value)

        if self.editing:
            name = self.request.matchdict["name"]
            query = query.filter(model.Setting.name != name)

        if query.count():
            node.raise_invalid("Setting name must be unique")


def defaults(config, **kwargs):  # pylint: disable=missing-function-docstring
    base = globals()

    AppInfoView = kwargs.get(  # pylint: disable=invalid-name,redefined-outer-name
        "AppInfoView", base["AppInfoView"]
    )
    AppInfoView.defaults(config)

    SettingView = kwargs.get(  # pylint: disable=invalid-name,redefined-outer-name
        "SettingView", base["SettingView"]
    )
    SettingView.defaults(config)


def includeme(config):  # pylint: disable=missing-function-docstring
    defaults(config)



# -*- coding: utf-8; -*-

from wuttaweb import diffs as mod
from wuttaweb.testing import WebTestCase, VersionWebTestCase


# nb. using WebTestCase here only for mako support in render_html()
class TestDiff(WebTestCase):

    def make_diff(self, *args, **kwargs):
        return mod.Diff(*args, **kwargs)

    def test_constructor(self):
        old_data = {"foo": "bar"}
        new_data = {"foo": "baz"}
        diff = self.make_diff(old_data, new_data, fields=["foo"])
        self.assertEqual(diff.fields, ["foo"])

    def test_make_fields(self):
        old_data = {"foo": "bar"}
        new_data = {"foo": "bar", "baz": "zer"}
        # nb. this calls make_fields()
        diff = self.make_diff(old_data, new_data)
        # TODO: should the fields be cumulative? or just use new_data?
        self.assertEqual(diff.fields, ["baz", "foo"])

    def test_values(self):
        old_data = {"foo": "bar"}
        new_data = {"foo": "baz"}
        diff = self.make_diff(old_data, new_data)
        self.assertEqual(diff.old_value("foo"), "bar")
        self.assertEqual(diff.new_value("foo"), "baz")

    def test_values_differ(self):
        old_data = {"foo": "bar"}
        new_data = {"foo": "baz"}
        diff = self.make_diff(old_data, new_data)
        self.assertTrue(diff.values_differ("foo"))

        old_data = {"foo": "bar"}
        new_data = {"foo": "bar"}
        diff = self.make_diff(old_data, new_data)
        self.assertFalse(diff.values_differ("foo"))

    def test_render_values(self):
        old_data = {"foo": "bar"}
        new_data = {"foo": "baz"}
        diff = self.make_diff(old_data, new_data)
        self.assertEqual(diff.render_old_value("foo"), "'bar'")
        self.assertEqual(diff.render_new_value("foo"), "'baz'")

    def test_get_old_value_attrs(self):

        # no change
        old_data = {"foo": "bar"}
        new_data = {"foo": "bar"}
        diff = self.make_diff(old_data, new_data, nature="update")
        self.assertEqual(diff.get_old_value_attrs(False), {})

        # update
        old_data = {"foo": "bar"}
        new_data = {"foo": "baz"}
        diff = self.make_diff(old_data, new_data, nature="update")
        self.assertEqual(
            diff.get_old_value_attrs(True),
            {"style": f"background-color: {diff.old_color};"},
        )

        # delete
        old_data = {"foo": "bar"}
        new_data = {}
        diff = self.make_diff(old_data, new_data, nature="delete")
        self.assertEqual(
            diff.get_old_value_attrs(True),
            {"style": f"background-color: {diff.old_color};"},
        )

    def test_get_new_value_attrs(self):

        # no change
        old_data = {"foo": "bar"}
        new_data = {"foo": "bar"}
        diff = self.make_diff(old_data, new_data, nature="update")
        self.assertEqual(diff.get_new_value_attrs(False), {})

        # update
        old_data = {"foo": "bar"}
        new_data = {"foo": "baz"}
        diff = self.make_diff(old_data, new_data, nature="update")
        self.assertEqual(
            diff.get_new_value_attrs(True),
            {"style": f"background-color: {diff.new_color};"},
        )

        # create
        old_data = {}
        new_data = {"foo": "bar"}
        diff = self.make_diff(old_data, new_data, nature="create")
        self.assertEqual(
            diff.get_new_value_attrs(True),
            {"style": f"background-color: {diff.new_color};"},
        )

    def test_render_field_row(self):
        old_data = {"foo": "bar"}
        new_data = {"foo": "baz"}
        diff = self.make_diff(old_data, new_data)
        row = diff.render_field_row("foo")
        self.assertIn("<tr>", row)
        self.assertIn("&#39;bar&#39;", row)
        self.assertIn(f'style="background-color: {diff.old_color};"', row)
        self.assertIn("&#39;baz&#39;", row)
        self.assertIn(f'style="background-color: {diff.new_color};"', row)
        self.assertIn("</tr>", row)

    def test_render_html(self):
        old_data = {"foo": "bar"}
        new_data = {"foo": "baz"}
        diff = self.make_diff(old_data, new_data)
        html = diff.render_html()
        self.assertIn("<table", html)
        self.assertIn("<tr>", html)
        self.assertIn("&#39;bar&#39;", html)
        self.assertIn(f'style="background-color: {diff.old_color};"', html)
        self.assertIn("&#39;baz&#39;", html)
        self.assertIn(f'style="background-color: {diff.new_color};"', html)
        self.assertIn("</tr>", html)
        self.assertIn("</table>", html)


class TestVersionDiff(VersionWebTestCase):

    def make_diff(self, *args, **kwargs):
        return mod.VersionDiff(*args, **kwargs)

    def test_constructor(self):
        import sqlalchemy_continuum as continuum

        model = self.app.model
        user = model.User(username="fred")
        self.session.add(user)
        self.session.commit()
        user.username = "freddie"
        self.session.commit()
        self.session.delete(user)
        self.session.commit()

        txncls = continuum.transaction_class(model.User)
        vercls = continuum.version_class(model.User)
        versions = self.session.query(vercls).order_by(vercls.transaction_id).all()
        self.assertEqual(len(versions), 3)

        version = versions[0]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "create")
        self.assertEqual(
            diff.fields,
            ["active", "person_uuid", "prevent_edit", "username", "uuid"],
        )

        version = versions[1]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "update")
        self.assertEqual(
            diff.fields,
            ["active", "person_uuid", "prevent_edit", "username", "uuid"],
        )

        version = versions[2]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "delete")
        self.assertEqual(
            diff.fields,
            ["active", "person_uuid", "prevent_edit", "username", "uuid"],
        )

    def test_render_values(self):
        import sqlalchemy_continuum as continuum

        model = self.app.model
        user = model.User(username="fred")
        self.session.add(user)
        self.session.commit()
        user.username = "freddie"
        self.session.commit()
        self.session.delete(user)
        self.session.commit()

        txncls = continuum.transaction_class(model.User)
        vercls = continuum.version_class(model.User)
        versions = self.session.query(vercls).order_by(vercls.transaction_id).all()
        self.assertEqual(len(versions), 3)

        version = versions[0]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "create")
        self.assertEqual(diff.render_old_value("username"), "")
        self.assertEqual(
            diff.render_new_value("username"),
            '<span style="font-family: monospace;">&#39;fred&#39;</span>',
        )

        version = versions[1]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "update")
        self.assertEqual(
            diff.render_old_value("username"),
            '<span style="font-family: monospace;">&#39;fred&#39;</span>',
        )
        self.assertEqual(
            diff.render_new_value("username"),
            '<span style="font-family: monospace;">&#39;freddie&#39;</span>',
        )

        version = versions[2]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "delete")
        self.assertEqual(
            diff.render_old_value("username"),
            '<span style="font-family: monospace;">&#39;freddie&#39;</span>',
        )
        self.assertEqual(diff.render_new_value("username"), "")



# -*- coding: utf-8; -*-

from unittest.mock import patch

from sqlalchemy import orm

import colander

from wuttaweb.grids import Grid
from wuttaweb.views import users as mod
from wuttaweb.testing import WebTestCase


class TestUserView(WebTestCase):

    def make_view(self):
        return mod.UserView(self.request)

    def test_includeme(self):
        self.pyramid_config.include("wuttaweb.views.users")

    def test_get_query(self):
        view = self.make_view()
        query = view.get_query(session=self.session)
        self.assertIsInstance(query, orm.Query)

    def test_configure_grid(self):
        model = self.app.model
        view = self.make_view()
        grid = view.make_grid(model_class=model.User)
        self.assertFalse(grid.is_linked("person"))
        view.configure_grid(grid)
        self.assertTrue(grid.is_linked("person"))

    def test_grid_row_class(self):
        model = self.app.model
        user = model.User(username="barney", active=True)
        data = dict(user)
        view = self.make_view()

        self.assertIsNone(view.grid_row_class(user, data, 1))

        user.active = False
        self.assertEqual(view.grid_row_class(user, data, 1), "has-background-warning")

    def test_is_editable(self):
        model = self.app.model
        view = self.make_view()

        # active user is editable
        user = model.User(username="barney", active=True)
        self.assertTrue(view.is_editable(user))

        # inactive also editable
        user = model.User(username="barney", active=False)
        self.assertTrue(view.is_editable(user))

        # but not if prevent_edit flag is set
        user = model.User(username="barney", prevent_edit=True)
        self.assertFalse(view.is_editable(user))

        # unless request user is root
        self.request.is_root = True
        self.assertTrue(view.is_editable(user))

    def test_configure_form(self):
        model = self.app.model
        person = model.Person(
            first_name="Barney", last_name="Rubble", full_name="Barney Rubble"
        )
        barney = model.User(username="barney", person=person)
        self.session.add(barney)
        self.session.commit()
        view = self.make_view()

        # person replaced with first/last name when creating or editing
        with patch.object(view, "viewing", new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn("person", form)
            self.assertNotIn("first_name", form)
            self.assertNotIn("last_name", form)
            view.configure_form(form)
            self.assertIn("person", form)
            self.assertNotIn("first_name", form)
            self.assertNotIn("last_name", form)
        with patch.object(view, "creating", new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn("person", form)
            self.assertNotIn("first_name", form)
            self.assertNotIn("last_name", form)
            view.configure_form(form)
            self.assertNotIn("person", form)
            self.assertIn("first_name", form)
            self.assertIn("last_name", form)
        with patch.object(view, "editing", new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn("person", form)
            self.assertNotIn("first_name", form)
            self.assertNotIn("last_name", form)
            view.configure_form(form)
            self.assertNotIn("person", form)
            self.assertIn("first_name", form)
            self.assertIn("last_name", form)

        # first/last name have default values when editing
        with patch.object(view, "editing", new=True):
            form = view.make_form(model_instance=barney)
            self.assertNotIn("first_name", form.defaults)
            self.assertNotIn("last_name", form.defaults)
            view.configure_form(form)
            self.assertIn("first_name", form.defaults)
            self.assertEqual(form.defaults["first_name"], "Barney")
            self.assertIn("last_name", form.defaults)
            self.assertEqual(form.defaults["last_name"], "Rubble")

        # password removed (always, for now)
        with patch.object(view, "viewing", new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn("password", form)
            view.configure_form(form)
            self.assertNotIn("password", form)
        with patch.object(view, "editing", new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn("password", form)
            view.configure_form(form)
            self.assertNotIn("password", form)

        # api tokens grid shown only if current user has perm
        with patch.object(view, "viewing", new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn("api_tokens", form)
            view.configure_form(form)
            self.assertNotIn("api_tokens", form)
            with patch.object(self.request, "is_root", new=True):
                form = view.make_form(model_instance=barney)
                self.assertIn("api_tokens", form)
                view.configure_form(form)
                self.assertIn("api_tokens", form)

    def test_unique_username(self):
        model = self.app.model
        view = self.make_view()

        user = model.User(username="foo")
        self.session.add(user)
        self.session.commit()

        with patch.object(view, "Session", return_value=self.session):

            # invalid if same username in data
            node = colander.SchemaNode(colander.String(), name="username")
            self.assertRaises(colander.Invalid, view.unique_username, node, "foo")

            # but not if username belongs to current user
            view.editing = True
            self.request.matchdict = {"uuid": user.uuid}
            node = colander.SchemaNode(colander.String(), name="username")
            self.assertIsNone(view.unique_username(node, "foo"))

    def test_objectify(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        view = self.make_view()

        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        others = model.Role(name="Others")
        self.session.add(others)
        barney = model.User(username="barney")
        auth.set_user_password(barney, "testpass")
        barney.roles.append(blokes)
        self.session.add(barney)
        self.session.commit()

        with patch.object(self.request, "matchdict", new={"uuid": barney.uuid}):
            with patch.object(view, "editing", new=True):

                # sanity check, user is just in 'blokes' role
                self.session.refresh(barney)
                self.assertEqual(len(barney.roles), 1)
                self.assertEqual(barney.roles[0].name, "Blokes")

                # form can update user password
                self.assertTrue(auth.check_user_password(barney, "testpass"))
                form = view.make_model_form(model_instance=barney)
                form.validated = {"username": "barney", "set_password": "testpass2"}
                with patch.object(view, "Session", return_value=self.session):
                    user = view.objectify(form)
                self.assertIs(user, barney)
                self.assertTrue(auth.check_user_password(barney, "testpass2"))

                # form can update user roles
                form = view.make_model_form(model_instance=barney)
                form.validated = {"username": "barney", "roles": {others.uuid}}
                with patch.object(view, "Session", return_value=self.session):
                    user = view.objectify(form)
                self.assertIs(user, barney)
                self.assertEqual(len(user.roles), 1)
                self.assertEqual(user.roles[0].name, "Others")

                # person is auto-created
                self.assertIsNone(barney.person)
                form = view.make_model_form(model_instance=barney)
                form.validated = {
                    "username": "barney",
                    "first_name": "Barney",
                    "last_name": "Rubble",
                }
                with patch.object(view, "Session", return_value=self.session):
                    user = view.objectify(form)
                self.assertIsNotNone(barney.person)
                self.assertEqual(barney.person.first_name, "Barney")
                self.assertEqual(barney.person.last_name, "Rubble")
                self.assertEqual(barney.person.full_name, "Barney Rubble")

                # person is auto-removed
                self.assertIsNotNone(barney.person)
                form = view.make_model_form(model_instance=barney)
                form.validated = {
                    "username": "barney",
                    "first_name": "",
                    "last_name": "",
                }
                with patch.object(view, "Session", return_value=self.session):
                    user = view.objectify(form)
                self.assertIsNone(barney.person)

                # nb. re-attach the person
                barney.person = self.session.query(model.Person).one()

                # person name is updated
                self.assertEqual(barney.person.first_name, "Barney")
                self.assertEqual(barney.person.last_name, "Rubble")
                self.assertEqual(barney.person.full_name, "Barney Rubble")
                form = view.make_model_form(model_instance=barney)
                form.validated = {
                    "username": "barney",
                    "first_name": "Fred",
                    "last_name": "Flintstone",
                }
                with patch.object(view, "Session", return_value=self.session):
                    user = view.objectify(form)
                self.assertIsNotNone(barney.person)
                self.assertEqual(barney.person.first_name, "Fred")
                self.assertEqual(barney.person.last_name, "Flintstone")
                self.assertEqual(barney.person.full_name, "Fred Flintstone")

        with patch.object(view, "creating", new=True):

            # person is auto-created when making new user
            form = view.make_model_form()
            form.validated = {
                "username": "betty",
                "first_name": "Betty",
                "last_name": "Boop",
            }
            with patch.object(view, "Session", return_value=self.session):
                user = view.objectify(form)
            self.assertIsNotNone(user.person)
            self.assertEqual(user.person.first_name, "Betty")
            self.assertEqual(user.person.last_name, "Boop")
            self.assertEqual(user.person.full_name, "Betty Boop")

            # nb. keep ref to last user
            last_user = user

            # person is *not* auto-created if no name provided
            form = view.make_model_form()
            form.validated = {"username": "betty", "first_name": "", "last_name": ""}
            with patch.object(view, "Session", return_value=self.session):
                user = view.objectify(form)
            self.assertIsNone(user.person)
            self.assertIsNot(user, last_user)

    def test_update_roles(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        admin = auth.get_role_administrator(self.session)
        authed = auth.get_role_authenticated(self.session)
        anon = auth.get_role_anonymous(self.session)
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        others = model.Role(name="Others")
        self.session.add(others)
        barney = model.User(username="barney")
        barney.roles.append(blokes)
        self.session.add(barney)
        self.session.commit()
        view = self.make_view()
        view.editing = True
        self.request.matchdict = {"uuid": barney.uuid}

        # no error if data is missing roles
        form = view.make_model_form(model_instance=barney)
        form.validated = {"username": "barneyx"}
        user = view.objectify(form)
        self.assertIs(user, barney)
        self.assertEqual(barney.username, "barneyx")

        # sanity check, user is just in 'blokes' role
        self.session.refresh(barney)
        self.assertEqual(len(barney.roles), 1)
        self.assertEqual(barney.roles[0].name, "Blokes")

        # let's test a bunch at once to ensure:
        # - user roles are updated
        # - authed / anon roles are not added
        # - admin role not added if current user is not root
        form = view.make_model_form(model_instance=barney)
        form.validated = {
            "username": "barney",
            "roles": {admin.uuid, authed.uuid, anon.uuid, others.uuid},
        }
        with patch.object(view, "Session", return_value=self.session):
            user = view.objectify(form)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 1)
        self.assertEqual(user.roles[0].name, "Others")

        # let's test a bunch at once to ensure:
        # - user roles are updated
        # - admin role is added if current user is root
        self.request.is_root = True
        form = view.make_model_form(model_instance=barney)
        form.validated = {
            "username": "barney",
            "roles": {admin.uuid, blokes.uuid, others.uuid},
        }
        with patch.object(view, "Session", return_value=self.session):
            user = view.objectify(form)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 3)
        role_uuids = set([role.uuid for role in user.roles])
        self.assertEqual(role_uuids, {admin.uuid, blokes.uuid, others.uuid})

        # admin role not removed if current user is not root
        self.request.is_root = False
        form = view.make_model_form(model_instance=barney)
        form.validated = {"username": "barney", "roles": {blokes.uuid, others.uuid}}
        with patch.object(view, "Session", return_value=self.session):
            user = view.objectify(form)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 3)

        # admin role is removed if current user is root
        self.request.is_root = True
        form = view.make_model_form(model_instance=barney)
        form.validated = {"username": "barney", "roles": {blokes.uuid, others.uuid}}
        with patch.object(view, "Session", return_value=self.session):
            user = view.objectify(form)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 2)

    def test_normalize_api_token(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        view = self.make_view()

        user = model.User(username="foo")
        self.session.add(user)
        token = auth.add_api_token(user, "test token")
        self.session.commit()

        normal = view.normalize_api_token(token)
        self.assertIn("uuid", normal)
        self.assertEqual(normal["uuid"], token.uuid.hex)
        self.assertIn("description", normal)
        self.assertEqual(normal["description"], "test token")
        self.assertIn("created", normal)

    def test_make_api_tokens_grid(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        view = self.make_view()

        user = model.User(username="foo")
        self.session.add(user)
        token1 = auth.add_api_token(user, "test1")
        token2 = auth.add_api_token(user, "test2")
        self.session.commit()

        # grid should have 2 records but no tools/actions
        grid = view.make_api_tokens_grid(user)
        self.assertIsInstance(grid, Grid)
        self.assertEqual(len(grid.data), 2)
        self.assertEqual(len(grid.tools), 0)
        self.assertEqual(len(grid.actions), 0)

        # create + delete allowed
        with patch.object(self.request, "is_root", new=True):
            grid = view.make_api_tokens_grid(user)
            self.assertEqual(len(grid.tools), 1)
            self.assertIn("create", grid.tools)
            self.assertEqual(len(grid.actions), 1)
            self.assertEqual(grid.actions[0].key, "delete")

    def test_add_api_token(self):
        model = self.app.model
        view = self.make_view()

        user = model.User(username="foo")
        self.session.add(user)
        self.session.commit()
        self.session.refresh(user)
        self.assertEqual(len(user.api_tokens), 0)

        with patch.object(view, "Session", return_value=self.session):
            with patch.object(self.request, "matchdict", new={"uuid": user.uuid}):
                with patch.object(
                    self.request,
                    "json_body",
                    create=True,
                    new={"description": "testing"},
                ):
                    result = view.add_api_token()
                    self.assertEqual(len(user.api_tokens), 1)
                    token = user.api_tokens[0]
                    self.assertEqual(result["token_string"], token.token_string)
                    self.assertEqual(result["description"], "testing")

    def test_delete_api_token(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        view = self.make_view()

        user = model.User(username="foo")
        self.session.add(user)
        token1 = auth.add_api_token(user, "test1")
        token2 = auth.add_api_token(user, "test2")
        self.session.commit()
        self.session.refresh(user)
        self.assertEqual(len(user.api_tokens), 2)

        with patch.object(view, "Session", return_value=self.session):
            with patch.object(self.request, "matchdict", new={"uuid": user.uuid}):

                # normal behavior
                with patch.object(
                    self.request,
                    "json_body",
                    create=True,
                    new={"uuid": token1.uuid.hex},
                ):
                    result = view.delete_api_token()
                    self.assertEqual(result, {})
                    self.session.refresh(user)
                    self.assertEqual(len(user.api_tokens), 1)
                    token = user.api_tokens[0]
                    self.assertIs(token, token2)

                # token for wrong user
                user2 = model.User(username="bar")
                self.session.add(user2)
                token3 = auth.add_api_token(user2, "test3")
                self.session.commit()
                with patch.object(
                    self.request,
                    "json_body",
                    create=True,
                    new={"uuid": token3.uuid.hex},
                ):
                    result = view.delete_api_token()
                    self.assertEqual(result, {"error": "API token not found"})

                # token not found
                with patch.object(
                    self.request,
                    "json_body",
                    create=True,
                    new={"uuid": self.app.make_true_uuid().hex},
                ):
                    result = view.delete_api_token()
                    self.assertEqual(result, {"error": "API token not found"})



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for users
"""

from wuttjamaican.db.model import User
from wuttaweb.views import MasterView
from wuttaweb.forms import widgets
from wuttaweb.forms.schema import PersonRef, RoleRefs


class UserView(MasterView):  # pylint: disable=abstract-method
    """
    Master view for users.

    Default route prefix is ``users``.

    Notable URLs provided by this class:

    * ``/users/``
    * ``/users/new``
    * ``/users/XXX``
    * ``/users/XXX/edit``
    * ``/users/XXX/delete``
    """

    model_class = User

    labels = {
        "api_tokens": "API Tokens",
    }

    grid_columns = [
        "username",
        "person",
        "active",
    ]

    filter_defaults = {
        "username": {"active": True},
        "active": {"active": True, "verb": "is_true"},
    }
    sort_defaults = "username"

    form_fields = [
        "username",
        "person",
        "active",
        "prevent_edit",
        "roles",
        "api_tokens",
    ]

    def get_query(self, session=None):  # pylint: disable=empty-docstring
        """ """
        query = super().get_query(session=session)

        # nb. always join Person
        model = self.app.model
        query = query.outerjoin(model.Person)

        return query

    def configure_grid(self, grid):  # pylint: disable=empty-docstring
        """ """
        g = grid
        super().configure_grid(g)
        model = self.app.model

        # never show these
        g.remove("person_uuid", "role_refs", "password")

        # username
        g.set_link("username")

        # person
        g.set_link("person")
        g.set_sorter("person", model.Person.full_name)
        g.set_filter("person", model.Person.full_name, label="Person Full Name")

    def grid_row_class(  # pylint: disable=empty-docstring,unused-argument
        self, user, data, i
    ):
        """ """
        if not user.active:
            return "has-background-warning"
        return None

    def is_editable(self, obj):  # pylint: disable=empty-docstring
        """ """
        user = obj

        # only root can edit certain users
        if user.prevent_edit and not self.request.is_root:
            return False

        return True

    def configure_form(self, form):  # pylint: disable=empty-docstring
        """ """
        f = form
        super().configure_form(f)
        user = f.model_instance

        # username
        f.set_validator("username", self.unique_username)

        # person
        if self.creating or self.editing:
            f.fields.insert_after("person", "first_name")
            f.set_required("first_name", False)
            f.fields.insert_after("first_name", "last_name")
            f.set_required("last_name", False)
            f.remove("person")
            if self.editing:
                person = user.person
                if person:
                    f.set_default("first_name", person.first_name)
                    f.set_default("last_name", person.last_name)
        else:
            f.set_node("person", PersonRef(self.request))

        # password
        # nb. we must avoid 'password' as field name since
        # ColanderAlchemy wants to handle the raw/hashed value
        f.remove("password")
        # nb. no need for password field if readonly
        if self.creating or self.editing:
            # nb. use 'set_password' as field name
            f.append("set_password")
            f.set_required("set_password", False)
            f.set_widget("set_password", widgets.CheckedPasswordWidget())

        # roles
        f.append("roles")
        f.set_node("roles", RoleRefs(self.request))
        if not self.creating:
            f.set_default("roles", [role.uuid.hex for role in user.roles])

        # api_tokens
        if self.viewing and self.has_perm("manage_api_tokens"):
            f.set_grid("api_tokens", self.make_api_tokens_grid(user))
        else:
            f.remove("api_tokens")

    def unique_username(self, node, value):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        session = self.Session()

        query = session.query(model.User).filter(model.User.username == value)

        if self.editing:
            uuid = self.request.matchdict["uuid"]
            query = query.filter(model.User.uuid != uuid)

        if query.count():
            node.raise_invalid("Username must be unique")

    def objectify(self, form):  # pylint: disable=empty-docstring
        """ """
        auth = self.app.get_auth_handler()
        data = form.validated

        # normal logic first
        user = super().objectify(form)

        # maybe update person name
        if "first_name" in form or "last_name" in form:
            first_name = data.get("first_name")
            last_name = data.get("last_name")
            if self.creating and (first_name or last_name):
                user.person = auth.make_person(
                    first_name=first_name, last_name=last_name
                )
            elif self.editing:
                if first_name or last_name:
                    if user.person:
                        person = user.person
                        if "first_name" in form:
                            person.first_name = first_name
                        if "last_name" in form:
                            person.last_name = last_name
                        person.full_name = self.app.make_full_name(
                            person.first_name, person.last_name
                        )
                    else:
                        user.person = auth.make_person(
                            first_name=first_name, last_name=last_name
                        )
                elif user.person:
                    user.person = None

        # maybe set user password
        if "set_password" in form and data.get("set_password"):
            auth.set_user_password(user, data["set_password"])

        # update roles for user
        # TODO
        # if self.has_perm('edit_roles'):
        self.update_roles(user, form)

        return user

    def update_roles(self, user, form):  # pylint: disable=empty-docstring
        """ """
        # TODO
        # if not self.has_perm('edit_roles'):
        #     return
        data = form.validated
        if "roles" not in data:
            return

        model = self.app.model
        session = self.Session()
        auth = self.app.get_auth_handler()

        old_roles = {role.uuid for role in user.roles}
        new_roles = data["roles"]

        admin = auth.get_role_administrator(session)
        ignored = {
            auth.get_role_authenticated(session).uuid,
            auth.get_role_anonymous(session).uuid,
        }

        # add any new roles for the user, taking care to avoid certain
        # unwanted operations for built-in roles
        for uuid in new_roles:
            if uuid in ignored:
                continue
            if uuid in old_roles:
                continue
            if uuid == admin.uuid and not self.request.is_root:
                continue
            role = session.get(model.Role, uuid)
            user.roles.append(role)

        # remove any roles which were *not* specified, taking care to
        # avoid certain unwanted operations for built-in roles
        for uuid in old_roles:
            if uuid in new_roles:
                continue
            if uuid == admin.uuid and not self.request.is_root:
                continue
            role = session.get(model.Role, uuid)
            user.roles.remove(role)

    def make_api_tokens_grid(self, user):
        """
        Make and return the grid for the API Tokens field.

        This is only shown when current user has permission to manage
        API tokens for other users.

        :rtype: :class:`~wuttaweb.grids.base.Grid`
        """
        route_prefix = self.get_route_prefix()

        grid = self.make_grid(
            key=f"{route_prefix}.view.api_tokens",
            data=[self.normalize_api_token(t) for t in user.api_tokens],
            columns=[
                "description",
                "created",
            ],
            sortable=True,
            sort_on_backend=False,
            sort_defaults=[("created", "desc")],
        )

        if self.has_perm("manage_api_tokens"):

            # create token
            button = self.make_button(
                "New",
                primary=True,
                icon_left="plus",
                **{"@click": "$emit('new-token')"},
            )
            grid.add_tool(button, key="create")

            # delete token
            grid.add_action(
                "delete",
                url="#",
                icon="trash",
                link_class="has-text-danger",
                click_handler="$emit('delete-token', props.row)",
            )

        return grid

    def normalize_api_token(self, token):  # pylint: disable=empty-docstring
        """ """
        return {
            "uuid": token.uuid.hex,
            "description": token.description,
            "created": self.app.render_datetime(token.created),
        }

    def add_api_token(self):
        """
        AJAX view for adding a new user API token.

        This calls
        :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.add_api_token()`
        for the creation logic.
        """
        session = self.Session()
        auth = self.app.get_auth_handler()
        user = self.get_instance()
        data = self.request.json_body

        token = auth.add_api_token(user, data["description"])
        session.flush()
        session.refresh(token)

        result = self.normalize_api_token(token)
        result["token_string"] = token.token_string
        result["_action_url_delete"] = "#"
        return result

    def delete_api_token(self):
        """
        AJAX view for deleting a user API token.

        This calls
        :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.delete_api_token()`
        for the deletion logic.
        """
        model = self.app.model
        session = self.Session()
        auth = self.app.get_auth_handler()
        user = self.get_instance()
        data = self.request.json_body

        token = session.get(model.UserAPIToken, data["uuid"])
        if not token:
            return {"error": "API token not found"}

        if token.user is not user:
            return {"error": "API token not found"}

        auth.delete_api_token(token)
        return {}

    @classmethod
    def defaults(cls, config):  # pylint: disable=empty-docstring
        """ """

        # nb. User may come from custom model
        wutta_config = config.registry.settings["wutta_config"]
        app = wutta_config.get_app()
        cls.model_class = app.model.User

        cls._user_defaults(config)
        cls._defaults(config)

    @classmethod
    def _user_defaults(cls, config):
        """
        Provide extra default configuration for the User master view.
        """
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        instance_url_prefix = cls.get_instance_url_prefix()
        model_title = cls.get_model_title()

        # manage API tokens
        config.add_wutta_permission(
            permission_prefix,
            f"{permission_prefix}.manage_api_tokens",
            f"Manage API tokens for any {model_title}",
        )
        config.add_route(
            f"{route_prefix}.add_api_token",
            f"{instance_url_prefix}/add-api-token",
            request_method="POST",
        )
        config.add_view(
            cls,
            attr="add_api_token",
            route_name=f"{route_prefix}.add_api_token",
            permission=f"{permission_prefix}.manage_api_tokens",
            renderer="json",
        )
        config.add_route(
            f"{route_prefix}.delete_api_token",
            f"{instance_url_prefix}/delete-api-token",
            request_method="POST",
        )
        config.add_view(
            cls,
            attr="delete_api_token",
            route_name=f"{route_prefix}.delete_api_token",
            permission=f"{permission_prefix}.manage_api_tokens",
            renderer="json",
        )


def defaults(config, **kwargs):  # pylint: disable=missing-function-docstring
    base = globals()

    UserView = kwargs.get(  # pylint: disable=invalid-name,redefined-outer-name
        "UserView", base["UserView"]
    )
    UserView.defaults(config)


def includeme(config):  # pylint: disable=missing-function-docstring
    defaults(config)



# -*- coding: utf-8; -*-

import datetime
import decimal
from unittest.mock import patch

import colander
import deform
from pyramid import testing

from wuttjamaican.util import get_timezone_by_name
from wuttaweb import grids
from wuttaweb.forms import widgets as mod
from wuttaweb.forms import schema
from wuttaweb.forms.schema import (
    FileDownload,
    PersonRef,
    RoleRefs,
    Permissions,
    WuttaDateTime,
    EmailRecipients,
)
from wuttaweb.testing import WebTestCase


class TestObjectRefWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault("renderer", deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def make_widget(self, **kwargs):
        return mod.ObjectRefWidget(self.request, **kwargs)

    def test_serialize(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()

        with patch.object(schema, "Session", return_value=self.session):

            # standard (editable)
            node = colander.SchemaNode(PersonRef(self.request))
            widget = self.make_widget()
            field = self.make_field(node)
            html = widget.serialize(field, person.uuid)
            self.assertIn("<b-select ", html)

            # readonly
            node = colander.SchemaNode(PersonRef(self.request))
            node.model_instance = person
            widget = self.make_widget()
            field = self.make_field(node)
            html = widget.serialize(field, person.uuid, readonly=True)
            self.assertIn("Betty Boop", html)
            self.assertNotIn("<a", html)

            # with hyperlink
            node = colander.SchemaNode(PersonRef(self.request))
            node.model_instance = person
            widget = self.make_widget(url=lambda p: "/foo")
            field = self.make_field(node)
            html = widget.serialize(field, person.uuid, readonly=True)
            self.assertIn("Betty Boop", html)
            self.assertIn("<a", html)
            self.assertIn('href="/foo"', html)

    def test_get_template_values(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()

        with patch.object(schema, "Session", return_value=self.session):

            # standard
            node = colander.SchemaNode(PersonRef(self.request))
            widget = self.make_widget()
            field = self.make_field(node)
            values = widget.get_template_values(field, person.uuid, {})
            self.assertIn("cstruct", values)
            self.assertNotIn("url", values)

            # readonly w/ empty option
            node = colander.SchemaNode(
                PersonRef(self.request, empty_option=("_empty_", "(empty)"))
            )
            widget = self.make_widget(readonly=True, url=lambda obj: "/foo")
            field = self.make_field(node)
            values = widget.get_template_values(field, "_empty_", {})
            self.assertIn("cstruct", values)
            self.assertNotIn("url", values)


class TestWuttaDateWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault("renderer", deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def make_widget(self, **kwargs):
        return mod.WuttaDateWidget(self.request, **kwargs)

    def test_serialize(self):
        node = colander.SchemaNode(colander.Date())
        field = self.make_field(node)

        # first try normal date
        widget = self.make_widget()
        dt = datetime.date(2025, 1, 15)

        # editable widget has normal picker html
        result = widget.serialize(field, str(dt))
        self.assertIn("<wutta-datepicker", result)

        # readonly is rendered per app convention
        result = widget.serialize(field, str(dt), readonly=True)
        self.assertEqual(result, "2025-01-15")

        # now try again with datetime
        widget = self.make_widget()
        dt = datetime.datetime(2025, 1, 15, 8, 35)

        # editable widget has normal picker html
        result = widget.serialize(field, str(dt))
        self.assertIn("<wutta-datepicker", result)

        # readonly is rendered per app convention
        result = widget.serialize(field, str(dt), readonly=True)
        self.assertEqual(result, "2025-01-15")


class TestWuttaDateTimeWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault("renderer", deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def make_widget(self, **kwargs):
        return mod.WuttaDateTimeWidget(self.request, **kwargs)

    def test_serialize_editable(self):
        tzlocal = get_timezone_by_name("America/New_York")
        with patch.object(self.app, "get_timezone", return_value=tzlocal):
            widget = self.make_widget()
            self.assertFalse(widget.readonly)
            node = colander.SchemaNode(WuttaDateTime(), widget=widget)
            field = self.make_field(node)

            # nb. input data (from schema type) is always "local, zone-aware, isoformat"
            dt = datetime.datetime(2024, 12, 12, 13, 49, tzinfo=tzlocal)
            result = widget.serialize(field, dt.isoformat())
            self.assertIn("<wutta-datepicker", result)

    def test_serialize_readonly(self):
        tzlocal = get_timezone_by_name("America/New_York")
        with patch.object(self.app, "get_timezone", return_value=tzlocal):
            widget = self.make_widget(readonly=True)
            self.assertTrue(widget.readonly)
            node = colander.SchemaNode(WuttaDateTime(), widget=widget)
            field = self.make_field(node)

            # null
            self.assertEqual(widget.serialize(field, colander.null), "")
            self.assertEqual(widget.serialize(field, None), "")
            self.assertEqual(widget.serialize(field, ""), "")

            # input data (from schema type) is always "local, zone-aware, isoformat"
            dt = datetime.datetime(2024, 12, 12, 13, 49, tzinfo=tzlocal)
            result = widget.serialize(field, dt.isoformat())
            self.assertEqual(result, "2024-12-12 13:49-0500")


class TestWuttaMoneyInputWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault("renderer", deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def make_widget(self, **kwargs):
        return mod.WuttaMoneyInputWidget(self.request, **kwargs)

    def test_serialize(self):
        node = colander.SchemaNode(schema.WuttaMoney(self.request))
        field = self.make_field(node)
        widget = self.make_widget()
        amount = decimal.Decimal("12.34")

        # editable widget has normal text input
        result = widget.serialize(field, str(amount))
        self.assertIn("<b-input", result)

        # readonly is rendered per app convention
        result = widget.serialize(field, str(amount), readonly=True)
        self.assertEqual(result, "<span>$12.34</span>")

        # readonly w/ null value
        result = widget.serialize(field, None, readonly=True)
        self.assertEqual(result, "<span></span>")


class TestFileDownloadWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault("renderer", deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):

        # nb. we let the field construct the widget via our type
        # (nb. at first we do not provide a url)
        node = colander.SchemaNode(FileDownload(self.request))
        field = self.make_field(node)
        widget = field.widget

        # null value
        html = widget.serialize(field, None, readonly=True)
        self.assertNotIn("<a ", html)
        self.assertIn("<span>", html)

        # path to nonexistent file
        html = widget.serialize(field, "/this/path/does/not/exist", readonly=True)
        self.assertNotIn("<a ", html)
        self.assertIn("<span>", html)

        # path to actual file
        datfile = self.write_file("data.txt", "hello\n" * 1000)
        html = widget.serialize(field, datfile, readonly=True)
        self.assertNotIn("<a ", html)
        self.assertIn("<span>", html)
        self.assertIn("data.txt", html)
        self.assertIn("kB)", html)

        # path to file, w/ url
        node = colander.SchemaNode(FileDownload(self.request, url="/download/blarg"))
        field = self.make_field(node)
        widget = field.widget
        html = widget.serialize(field, datfile, readonly=True)
        self.assertNotIn("<span>", html)
        self.assertIn('<a href="/download/blarg">', html)
        self.assertIn("data.txt", html)
        self.assertIn("kB)", html)

        # nb. same readonly output even if we ask for editable
        html2 = widget.serialize(field, datfile, readonly=False)
        self.assertEqual(html2, html)


class TestGridWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault("renderer", deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):
        grid = grids.Grid(
            self.request,
            columns=["foo", "bar"],
            data=[{"foo": 1, "bar": 2}, {"foo": 3, "bar": 4}],
        )

        node = colander.SchemaNode(colander.String())
        widget = mod.GridWidget(self.request, grid)
        field = self.make_field(node)

        # readonly works okay
        html = widget.serialize(field, None, readonly=True)
        self.assertIn("<b-table ", html)

        # but otherwise, error
        self.assertRaises(NotImplementedError, widget.serialize, field, None)


class TestRoleRefsWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault("renderer", deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):
        self.pyramid_config.add_route("roles.view", "/roles/{uuid}")
        model = self.app.model
        auth = self.app.get_auth_handler()
        admin = auth.get_role_administrator(self.session)
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        self.session.commit()

        # nb. we let the field construct the widget via our type
        with patch.object(schema, "Session", return_value=self.session):
            node = colander.SchemaNode(RoleRefs(self.request))
            field = self.make_field(node)
            widget = field.widget

            # readonly values list includes admin
            html = widget.serialize(field, {admin.uuid, blokes.uuid}, readonly=True)
            self.assertIn(admin.name, html)
            self.assertIn(blokes.name, html)

            # editable values list *excludes* admin (by default)
            html = widget.serialize(field, {admin.uuid, blokes.uuid})
            self.assertNotIn(str(admin.uuid.hex), html)
            self.assertIn(str(blokes.uuid.hex), html)

            # but admin is included for root user
            self.request.is_root = True
            node = colander.SchemaNode(RoleRefs(self.request))
            field = self.make_field(node)
            widget = field.widget
            html = widget.serialize(field, {admin.uuid, blokes.uuid})
            self.assertIn(str(admin.uuid.hex), html)
            self.assertIn(str(blokes.uuid.hex), html)


class TestPermissionsWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault("renderer", deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):
        permissions = {
            "widgets": {
                "label": "Widgets",
                "perms": {
                    "widgets.polish": {
                        "label": "Polish the widgets",
                    },
                },
            },
        }

        # nb. we let the field construct the widget via our type
        node = colander.SchemaNode(Permissions(self.request, permissions))
        field = self.make_field(node)
        widget = field.widget

        # readonly output does *not* include the perm by default
        html = widget.serialize(field, set(), readonly=True)
        self.assertNotIn("Polish the widgets", html)

        # readonly output includes the perm if set
        html = widget.serialize(field, {"widgets.polish"}, readonly=True)
        self.assertIn("Polish the widgets", html)

        # editable output always includes the perm
        html = widget.serialize(field, set())
        self.assertIn("Polish the widgets", html)


class TestEmailRecipientsWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault("renderer", deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):
        node = colander.SchemaNode(EmailRecipients())
        field = self.make_field(node)
        widget = mod.EmailRecipientsWidget()

        recips = [
            "alice@example.com",
            "bob@example.com",
        ]
        recips_str = ", ".join(recips)

        # readonly
        result = widget.serialize(field, recips_str, readonly=True)
        self.assertIn("<ul>", result)
        self.assertIn("<li>alice@example.com</li>", result)

        # editable
        result = widget.serialize(field, recips_str)
        self.assertIn("<b-input", result)
        self.assertIn('type="textarea"', result)

    def test_deserialize(self):
        node = colander.SchemaNode(EmailRecipients())
        field = self.make_field(node)
        widget = mod.EmailRecipientsWidget()

        recips = [
            "alice@example.com",
            "bob@example.com",
        ]
        recips_str = ", ".join(recips)

        # values
        result = widget.deserialize(field, recips_str)
        self.assertEqual(result, recips_str)

        # null
        result = widget.deserialize(field, colander.null)
        self.assertIs(result, colander.null)


class TestBatchIdWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault("renderer", deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):
        node = colander.SchemaNode(colander.Integer())
        field = self.make_field(node)
        widget = mod.BatchIdWidget()

        result = widget.serialize(field, colander.null)
        self.assertIs(result, colander.null)

        result = widget.serialize(field, 42)
        self.assertEqual(result, "00000042")



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Tools for displaying simple data diffs
"""

import sqlalchemy as sa

from pyramid.renderers import render
from webhelpers2.html import HTML


class Diff:
    """
    Represent / display a basic "diff" between two data records.

    You must provide both the "old" and "new" data records, when
    constructing an instance of this class.  Then call
    :meth:`render_html()` to display the diff table.

    :param old_data: Dict of "old" data record.

    :param new_data: Dict of "new" data record.

    :param fields: Optional list of field names.  If not specified,
       will be derived from the data records.

    :param nature: What sort of diff is being represented; must be one
       of: ``("create", "update", "delete")``

    :param old_color: Background color to display for "old/deleted"
       field data, when applicable.

    :param new_color: Background color to display for "new/created"
       field data, when applicable.
    """

    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        old_data: dict,
        new_data: dict,
        fields: list = None,
        nature="update",
        old_color="#ffebe9",
        new_color="#dafbe1",
    ):
        self.old_data = old_data
        self.new_data = new_data
        self.columns = ["field name", "old value", "new value"]
        self.fields = fields or self.make_fields()
        self.nature = nature
        self.old_color = old_color
        self.new_color = new_color

    def make_fields(self):  # pylint: disable=missing-function-docstring
        return sorted(set(self.old_data) | set(self.new_data), key=lambda x: x.lower())

    def old_value(self, field):  # pylint: disable=missing-function-docstring
        return self.old_data.get(field)

    def new_value(self, field):  # pylint: disable=missing-function-docstring
        return self.new_data.get(field)

    def values_differ(self, field):  # pylint: disable=missing-function-docstring
        return self.new_value(field) != self.old_value(field)

    def render_html(self, template="/diff.mako", **kwargs):
        """
        Render the diff as HTML table.

        :param template: Name of template to render, if you need to
           override the default.

        :param \\**kwargs: Remaining kwargs are passed as context to
           the template renderer.

        :returns: HTML literal string
        """
        context = kwargs
        context["diff"] = self
        return HTML.literal(render(template, context))

    def render_field_row(self, field):  # pylint: disable=missing-function-docstring
        is_diff = self.values_differ(field)

        td_field = HTML.tag("td", class_="field", c=field)

        td_old_value = HTML.tag(
            "td",
            c=self.render_old_value(field),
            **self.get_old_value_attrs(is_diff),
        )

        td_new_value = HTML.tag(
            "td",
            c=self.render_new_value(field),
            **self.get_new_value_attrs(is_diff),
        )

        return HTML.tag("tr", c=[td_field, td_old_value, td_new_value])

    def render_old_value(self, field):  # pylint: disable=missing-function-docstring
        value = self.old_value(field)
        return repr(value)

    def render_new_value(self, field):  # pylint: disable=missing-function-docstring
        value = self.new_value(field)
        return repr(value)

    def get_old_value_attrs(  # pylint: disable=missing-function-docstring
        self, is_diff
    ):
        attrs = {}
        if self.nature == "update" and is_diff:
            attrs["style"] = f"background-color: {self.old_color};"
        elif self.nature == "delete":
            attrs["style"] = f"background-color: {self.old_color};"
        return attrs

    def get_new_value_attrs(  # pylint: disable=missing-function-docstring
        self, is_diff
    ):
        attrs = {}
        if self.nature == "create":
            attrs["style"] = f"background-color: {self.new_color};"
        elif self.nature == "update" and is_diff:
            attrs["style"] = f"background-color: {self.new_color};"
        return attrs


class VersionDiff(Diff):
    """
    Special diff class, for use with version history views.  Note that
    while based on :class:`Diff`, this class uses a different
    signature for the constructor.

    :param version: Reference to a Continuum version record (object).

    :param \\**kwargs: Remaining kwargs are passed as-is to the
       :class:`Diff` constructor.
    """

    def __init__(self, version, **kwargs):
        import sqlalchemy_continuum as continuum  # pylint: disable=import-outside-toplevel
        from wutta_continuum.util import (  # pylint: disable=import-outside-toplevel
            render_operation_type,
        )

        self.version = version
        self.model_class = continuum.parent_class(type(self.version))
        self.mapper = sa.inspect(self.model_class)
        self.version_mapper = sa.inspect(type(self.version))
        self.title = kwargs.pop("title", self.model_class.__name__)

        self.operation_title = render_operation_type(self.version.operation_type)

        if "nature" not in kwargs:
            if (
                version.previous
                and version.operation_type == continuum.Operation.DELETE
            ):
                kwargs["nature"] = "delete"
            elif version.previous:
                kwargs["nature"] = "update"
            else:
                kwargs["nature"] = "create"

        if "fields" not in kwargs:
            kwargs["fields"] = self.get_default_fields()

        old_data = {}
        new_data = {}
        for field in kwargs["fields"]:
            if version.previous:
                old_data[field] = getattr(version.previous, field)
            new_data[field] = getattr(version, field)

        super().__init__(old_data, new_data, **kwargs)

    def get_default_fields(self):  # pylint: disable=missing-function-docstring
        fields = sorted(self.version_mapper.columns.keys())

        unwanted = [
            "transaction_id",
            "end_transaction_id",
            "operation_type",
        ]

        return [field for field in fields if field not in unwanted]

    def render_version_value(self, value):  # pylint: disable=missing-function-docstring
        return HTML.tag("span", c=[repr(value)], style="font-family: monospace;")

    def render_old_value(self, field):
        if self.nature == "create":
            return ""
        value = self.old_value(field)
        return self.render_version_value(value)

    def render_new_value(self, field):
        if self.nature == "delete":
            return ""
        value = self.new_value(field)
        return self.render_version_value(value)



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import patch, MagicMock

from wuttaweb import menus as mod
from wuttaweb.testing import WebTestCase


class TestMenuHandler(WebTestCase):

    def setUp(self):
        self.setup_web()
        self.handler = mod.MenuHandler(self.config)

    def test_make_admin_menu(self):

        # no people entry by default
        menu = self.handler.make_admin_menu(self.request)
        self.assertIsInstance(menu, dict)
        routes = [item.get("route") for item in menu["items"]]
        self.assertNotIn("people", routes)

        # but we can request it
        menu = self.handler.make_admin_menu(self.request, include_people=True)
        routes = [item.get("route") for item in menu["items"]]
        self.assertIn("people", routes)

    def test_make_menus(self):
        menus = self.handler.make_menus(self.request)
        self.assertIsInstance(menus, list)

    def test_is_allowed(self):
        model = self.app.model
        auth = self.app.get_auth_handler()

        # user with perms
        barney = model.User(username="barney")
        self.session.add(barney)
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        barney.roles.append(blokes)
        auth.grant_permission(blokes, "appinfo.list")
        self.request.user = barney

        # perm not granted to user
        item = {"perm": "appinfo.configure"}
        self.assertFalse(self.handler._is_allowed(self.request, item))

        # perm *is* granted to user
        item = {"perm": "appinfo.list"}
        self.assertTrue(self.handler._is_allowed(self.request, item))

        # perm not required
        item = {}
        self.assertTrue(self.handler._is_allowed(self.request, item))

    def test_mark_allowed(self):

        def make_menus():
            return [
                {
                    "type": "menu",
                    "items": [
                        {"title": "Foo", "url": "#"},
                        {"title": "Bar", "url": "#"},
                    ],
                },
            ]

        mock_is_allowed = MagicMock()
        with patch.object(self.handler, "_is_allowed", new=mock_is_allowed):

            # all should be allowed
            mock_is_allowed.return_value = True
            menus = make_menus()
            self.handler._mark_allowed(self.request, menus)
            menu = menus[0]
            self.assertTrue(menu["allowed"])
            foo, bar = menu["items"]
            self.assertTrue(foo["allowed"])
            self.assertTrue(bar["allowed"])

            # none should be allowed
            mock_is_allowed.return_value = False
            menus = make_menus()
            self.handler._mark_allowed(self.request, menus)
            menu = menus[0]
            self.assertFalse(menu["allowed"])
            foo, bar = menu["items"]
            self.assertFalse(foo["allowed"])
            self.assertFalse(bar["allowed"])

    def test_mark_allowed_submenu(self):

        def make_menus():
            return [
                {
                    "type": "menu",
                    "items": [
                        {"title": "Foo", "url": "#"},
                        {
                            "type": "menu",
                            "items": [
                                {"title": "Bar", "url": "#"},
                            ],
                        },
                    ],
                },
            ]

        mock_is_allowed = MagicMock()
        with patch.object(self.handler, "_is_allowed", new=mock_is_allowed):

            # all should be allowed
            mock_is_allowed.return_value = True
            menus = make_menus()
            self.handler._mark_allowed(self.request, menus)
            menu = menus[0]
            self.assertTrue(menu["allowed"])
            foo, submenu = menu["items"]
            self.assertTrue(foo["allowed"])
            self.assertTrue(submenu["allowed"])
            subitem = submenu["items"][0]
            self.assertTrue(subitem["allowed"])

            # none should be allowed
            mock_is_allowed.return_value = False
            menus = make_menus()
            self.handler._mark_allowed(self.request, menus)
            menu = menus[0]
            self.assertFalse(menu["allowed"])
            foo, submenu = menu["items"]
            self.assertFalse(foo["allowed"])
            self.assertFalse(submenu["allowed"])
            subitem = submenu["items"][0]
            self.assertFalse(subitem["allowed"])

    def test_make_menu_key(self):
        self.assertEqual(self.handler._make_menu_key("foo"), "foo")
        self.assertEqual(self.handler._make_menu_key("FooBar"), "foobar")
        self.assertEqual(self.handler._make_menu_key("Foo - $#Bar"), "foobar")
        self.assertEqual(self.handler._make_menu_key("Foo__Bar"), "foo__bar")

    def test_make_menu_entry_item(self):
        item = {"title": "Foo", "url": "#"}
        entry = self.handler._make_menu_entry(self.request, item)
        self.assertEqual(entry["type"], "item")
        self.assertEqual(entry["title"], "Foo")
        self.assertEqual(entry["url"], "#")
        self.assertTrue(entry["is_link"])

    def test_make_menu_entry_item_with_no_url(self):
        item = {"title": "Foo"}
        entry = self.handler._make_menu_entry(self.request, item)
        self.assertEqual(entry["type"], "item")
        self.assertEqual(entry["title"], "Foo")
        self.assertNotIn("url", entry)
        # nb. still sets is_link = True; basically it's <a> with no href
        self.assertTrue(entry["is_link"])

    def test_make_menu_entry_item_with_known_route(self):
        item = {"title": "Foo", "route": "home"}
        with patch.object(self.request, "route_url", return_value="/something"):
            entry = self.handler._make_menu_entry(self.request, item)
            self.assertEqual(entry["type"], "item")
            self.assertEqual(entry["url"], "/something")
            self.assertTrue(entry["is_link"])

    def test_make_menu_entry_item_with_unknown_route(self):
        item = {"title": "Foo", "route": "home"}
        with patch.object(self.request, "route_url", side_effect=KeyError):
            entry = self.handler._make_menu_entry(self.request, item)
            self.assertEqual(entry["type"], "item")
            # nb. fake url is used, based on (bad) route name
            self.assertEqual(entry["url"], "home")
            self.assertTrue(entry["is_link"])

    def test_make_menu_entry_sep(self):
        item = {"type": "sep"}
        entry = self.handler._make_menu_entry(self.request, item)
        self.assertEqual(entry["type"], "sep")
        self.assertTrue(entry["is_sep"])
        self.assertFalse(entry["is_menu"])

    def test_make_raw_menus(self):
        # minimal test to ensure it calls the other method
        with patch.object(self.handler, "make_menus") as make_menus:
            self.handler._make_raw_menus(self.request, foo="bar")
            make_menus.assert_called_once_with(self.request, foo="bar")

    def test_do_make_menus_prune_unallowed_item(self):
        test_menus = [
            {
                "title": "First Menu",
                "type": "menu",
                "items": [
                    {"title": "Foo", "url": "#"},
                    {"title": "Bar", "url": "#"},
                ],
            },
        ]

        def is_allowed(request, item):
            if item.get("title") == "Bar":
                return False
            return True

        with patch.object(self.handler, "make_menus", return_value=test_menus):
            with patch.object(self.handler, "_is_allowed", side_effect=is_allowed):
                menus = self.handler.do_make_menus(self.request)

                # Foo remains but Bar is pruned
                menu = menus[0]
                self.assertEqual(len(menu["items"]), 1)
                item = menu["items"][0]
                self.assertEqual(item["title"], "Foo")

    def test_do_make_menus_prune_unallowed_menu(self):
        test_menus = [
            {
                "title": "First Menu",
                "type": "menu",
                "items": [
                    {"title": "Foo", "url": "#"},
                    {"title": "Bar", "url": "#"},
                ],
            },
            {
                "title": "Second Menu",
                "type": "menu",
                "items": [
                    {"title": "Baz", "url": "#"},
                ],
            },
        ]

        def is_allowed(request, item):
            if item.get("title") == "Baz":
                return True
            return False

        with patch.object(self.handler, "make_menus", return_value=test_menus):
            with patch.object(self.handler, "_is_allowed", side_effect=is_allowed):
                menus = self.handler.do_make_menus(self.request)

                # Second/Baz remains but First/Foo/Bar are pruned
                self.assertEqual(len(menus), 1)
                menu = menus[0]
                self.assertEqual(menu["title"], "Second Menu")
                self.assertEqual(len(menu["items"]), 1)
                item = menu["items"][0]
                self.assertEqual(item["title"], "Baz")

    def test_do_make_menus_with_top_link(self):
        test_menus = [
            {
                "title": "First Menu",
                "type": "menu",
                "items": [
                    {"title": "Foo", "url": "#"},
                    {"title": "Bar", "url": "#"},
                ],
            },
            {
                "title": "Second Link",
                "type": "link",
            },
        ]

        with patch.object(self.handler, "make_menus", return_value=test_menus):
            with patch.object(self.handler, "_is_allowed", return_value=True):
                menus = self.handler.do_make_menus(self.request)

                # ensure top link remains
                self.assertEqual(len(menus), 2)
                menu = menus[1]
                self.assertEqual(menu["title"], "Second Link")

    def test_do_make_menus_with_trailing_sep(self):
        test_menus = [
            {
                "title": "First Menu",
                "type": "menu",
                "items": [
                    {"title": "Foo", "url": "#"},
                    {"title": "Bar", "url": "#"},
                    {"type": "sep"},
                ],
            },
        ]

        with patch.object(self.handler, "make_menus", return_value=test_menus):
            with patch.object(self.handler, "_is_allowed", return_value=True):
                menus = self.handler.do_make_menus(self.request)

                # ensure trailing sep was pruned
                menu = menus[0]
                self.assertEqual(len(menu["items"]), 2)
                foo, bar = menu["items"]
                self.assertEqual(foo["title"], "Foo")
                self.assertEqual(bar["title"], "Bar")

    def test_do_make_menus_with_submenu(self):
        test_menus = [
            {
                "title": "First Menu",
                "type": "menu",
                "items": [
                    {
                        "title": "First Submenu",
                        "type": "menu",
                        "items": [
                            {"title": "Foo", "url": "#"},
                        ],
                    },
                    {
                        "title": "Second Submenu",
                        "type": "menu",
                        "items": [
                            {"title": "Bar", "url": "#"},
                        ],
                    },
                ],
            },
        ]

        def is_allowed(request, item):
            if item.get("title") == "Bar":
                return False
            return True

        with patch.object(self.handler, "make_menus", return_value=test_menus):
            with patch.object(self.handler, "_is_allowed", side_effect=is_allowed):
                menus = self.handler.do_make_menus(self.request)

                # first submenu remains, second is pruned
                menu = menus[0]
                self.assertEqual(len(menu["items"]), 1)
                submenu = menu["items"][0]
                self.assertEqual(submenu["type"], "submenu")
                self.assertEqual(submenu["title"], "First Submenu")
                self.assertEqual(len(submenu["items"]), 1)
                item = submenu["items"][0]
                self.assertEqual(item["title"], "Foo")
